//***************************************************************************
//* CalcLuator: a simple Lua calculator plugin for Hiew. (c) FSC, 2007-2009 *
//*                                                                         *
//* This code expects the following directory structure:                    *
//*                                                                         *
//* HiewSDK                                                                 *
//*   HiewLuaSDK                                                            *
//*   SciViewSDK                                                            *
//* ThisCodeFolder                                                          *
//***************************************************************************

#include <windows.h>
#include "hem.h"
#include "SciView.h"
#include "HiewLua.h"
#include "Common.h"
#include "Scripter.h"
#include "Viewer.h"
#include "resource.h"

//------------------------------------------------------------------------
// Don't we want to be a (kilo)bit smaller?
//
#pragma comment(linker, "/OPT:NOWIN98")

//------------------------------------------------------------------------
// Local data
//
char SaveFileName[MAX_PATH];

HMENU hMainMenu = 0;

char HiewFileName[HEM_FILENAME_MAXLEN];
ULARGE_INTEGER HiewFileLength;
ULARGE_INTEGER HiewCurrentOffset;

ULARGE_INTEGER OutOffset;
HEM_DWORD OutFlags;
HEM_BYTE OutMode;

//------------------------------------------------------------------------
// Show an error message
//
int ShowError(LPCTSTR Message)
{
	HiewGate_Message((PHEM_BYTE)"CalcLuator Error", (PHEM_BYTE)Message);
	return HEM_OK;
}

//------------------------------------------------------------------------
// Plugin entry point
//
static int HEM_API Hem_EntryPoint(HEMCALL_TAG *hemCall)
{
	if(hemCall->cbSize < sizeof(HEMCALL_TAG)) return HEM_ERROR;

	HIEWGATE_GETDATA HiewData;
	if(HiewGate_GetData(&HiewData) != HEM_OK) return HEM_ERROR;

    /* remember important values that will be available to Lua scripts */
    strcpy(HiewFileName, (PCHAR)&HiewData.filename[0]);
    HiewFileLength.QuadPart = HiewData.filelength;
    HiewCurrentOffset.QuadPart = HiewData.offsetCurrent;

    /* load both necessary plugins */
    if(!LoadSciView()) return HEM_OK;
    if(!LoadHiewLua()) return HEM_OK;

    /* show SciView window */
    int RetVal = ShowSciView();
    if(RetVal == -1) ShowError("Could not create SciView window");

    if(RetVal == IDC_EXITTOHIEW)
    {
        hemCall->returnActionFlag = OutFlags;
        hemCall->returnMode = OutMode;
        hemCall->returnOffset = OutOffset.QuadPart;
    }

    /* it's not an error to call CloseLua() multiple times because HiewLua plugin checks its state */
    CloseLua();

    return HEM_OK;
}

//------------------------------------------------------------------------
// Plugin loading stuff
//
HEMINFO_TAG hemCalcLuator   =   {   sizeof(HEMINFO_TAG),
                                    sizeof(int),
                                    0,						// reserved
                                    HEM_SDK_VERSION_MAJOR, HEM_SDK_VERSION_MINOR,
                                    CALCLUATOR_VERSION_MAJOR, CALCLUATOR_VERSION_MINOR,
                                    HEM_FLAG_FILEMASK | HEM_FLAG_MODEMASK,
                                    0,						// reserved
                                    Hem_EntryPoint,
                                    NULL,                   // no Hem_Unload()
                                    NULL,                   // no Hem2HemGate()
                                    0,						// reserved
                                    0,						// reserved
                                    0,						// reserved
                                    0,						// reserved
                                    "CalcLuator",
                                    "Lua Calculator",
                                    "**********************************************",
                                    "* FSC (c) 2007-2009. http://www.fsc-soft.com *",
                                    "**********************************************"
                                };



int HEM_EXPORT Hem_Load(HIEWINFO_TAG *hiewInfo)
{
	/* check that SDK version is OK */
	DWORD SDKVersion = (hiewInfo->sdkVerMajor << 8) | hiewInfo->sdkVerMinor;
	DWORD RequiredSDKVersion = (HEM_SDK_VERSION_MAJOR << 8) | HEM_SDK_VERSION_MINOR;
	if(SDKVersion < RequiredSDKVersion) return HEM_ERR_SDKVER_INCOMPATIBLE;

	HiewGate_Set(hiewInfo);
	hiewInfo->hemInfo = &hemCalcLuator;
	return HEM_OK;
}

//------------------------------------------------------------------------
// DLL main entry point
//
BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch(fdwReason)
	{
		case DLL_PROCESS_ATTACH:
			DisableThreadLibraryCalls(hinstDLL);
            GetModuleFileName(hinstDLL, SaveFileName, sizeof(SaveFileName));
            PathRenameExtension(SaveFileName, ".sav");
            hMainMenu = LoadMenu(hinstDLL, MAKEINTRESOURCE(MAIN_MENU));
			break;

		case DLL_PROCESS_DETACH:
            if(hMainMenu) DestroyMenu(hMainMenu);
            hMainMenu = 0;
			break;
	}
	return TRUE;
}
