#include "hem.h"
#include "SciView.h"
#include "Common.h"
#include "Scripter.h"
#include "resource.h"
#include "HiewLua.h"

//------------------------------------------------------------------------
// Local definitions
//
#define STYLE_STDOUT    1
#define STYLE_STDERR    2

#define WM_SCRIPTDONE   WM_USER + 10000

#define DEVELOPERS_EMAIL    "fsc@lenta.ru"

//------------------------------------------------------------------------
// Forward declarations
//
static BOOL WmKeyDown(HWND hWnd, DWORD Vk);
void __stdcall PrintStdErr(LPCTSTR pText);

//------------------------------------------------------------------------
// Local data
//
static HIEWGATE_GETHEM2HEMGATE SciView;                         // gate to SciView plugin
static SCIVIEW_COMMAND SciViewCmd = {sizeof(SCIVIEW_COMMAND)};  // SciView never changes the cbSize member so its safe to initialize it only once

static SciFnDirect DirectFn = 0;
static sptr_t DirectPtr = 0;

HWND hDialogWnd = 0;
static HWND hSciViewWnd = 0;

//------------------------------------------------------------------------
// Call SciView
//
static int CallSciView(DWORD Cmd)
{
    SciViewCmd.Command = Cmd;
    return SciView.Hem2HemGate(&SciViewCmd);
}

//------------------------------------------------------------------------
// Send message to Scintilla
// Note: as far as I remember, SciView's FirectFn is not protected with a critical
// section so one should be very careful in multithreaded environment in order
// to avoid race conditions
//
static int SciSend(int Message, int wParam = 0, LPVOID lParam = 0)
{
    return DirectFn(DirectPtr, Message, wParam, (int)lParam);
}

//------------------------------------------------------------------------
// Close dialog with given exit code
//
static void CloseDialog(int ExitCode)
{
    SciViewCmd.IntValue = ExitCode;
    CallSciView(SCIVIEW_HIDE);
}

//------------------------------------------------------------------------
// Insert CRLF at the current position
//
static void CrLf()
{
    SciViewCmd.pText = "\r\n";
    CallSciView(SCIVIEW_ADDTEXT);
}

//------------------------------------------------------------------------
// Load SciView
//
bool LoadSciView()
{
    /* get SciView entry point */
    if(HEM_OK != HiewGate_GetHem2HemGate(&SciView, (PHEM_BYTE)"SciView") || SciView.Hem2HemGate == 0)
    {
        ShowError("Could not load SciView plugin");
        return false;
    }
    return true;
}

//------------------------------------------------------------------------
// Preview Scintilla messages
// Note that we won't see here any messages delivered through DirectFn/DirectPtr
// Also note that we are called BEFORE the actual processing of the message
// so if we change any message parameters (which is entirely OK), Scintilla window
// will see and process the new updated message
//
void __stdcall MessagePreview(HWND, UINT *pMsg, WPARAM *wParam, LPARAM *)
{
    switch(*pMsg)
    {
        case WM_KEYDOWN:
        case WM_SYSKEYDOWN:
            if(WmKeyDown(hDialogWnd, (DWORD)*wParam))
            {
                /* don't let Scintilla to see already processed keystrokes */
                *pMsg = WM_NULL;
            }
            break;

        case WM_CHAR:
            {
                /* suppress chars that are caused by known key combinations */
                char Ch = *wParam;
                switch(Ch)
                {
                    case 1:     // Ctrl+A
                    case 3:     // Ctrl+C, Ctrl+Break
                    case 4:     // Ctrl+D
                    case 6:     // Ctrl+F
                    case 9:     // Ctrl+I
                    case 0xc:   // Ctrl+L
                    case 0x10:  // Ctrl+P
                    case 0x12:  // Ctrl+R
                    case 0x17:  // Ctrl+W
                        *pMsg = 0;
                        break;
                }
            }
            break;
    }
}

//------------------------------------------------------------------------
// Load .sav file
// Returns true if .sav file has been loaded, false otherwise
// Note: no error checking here because the operation is not vital
//
static bool LoadSav()
{
    bool Loaded = false;
    if(PathFileExists(SaveFileName))
    {
        HANDLE hFile = CreateFile(SaveFileName, GENERIC_READ, FILE_SHARE_READ, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
        if(hFile != INVALID_HANDLE_VALUE)
        {
            DWORD FileSize = GetFileSize(hFile, 0);
            PBYTE pContents = (PBYTE)malloc(FileSize);
            if(pContents)
            {
                DWORD BytesRead;
                if(ReadFile(hFile, pContents, FileSize, &BytesRead, 0) && BytesRead == FileSize)
                {
                    SciSend(SCI_ADDSTYLEDTEXT, FileSize, pContents);
                    Loaded = true;
                }
                free(pContents);
            }
        }
        CloseHandle(hFile);
    }
    return Loaded;
}

//------------------------------------------------------------------------
// Setup SciView
// Note: we CANNOT properly setup SciView before its window is created.
// Therefore we call this routine from the WM_INITDIALOG message handler
//
static SCIVIEW_PARAM_SETUPSTYLE LuaStyles[] =
{
    {           0, SETUPSTYLE_ALL, "Courier", 10, 0, 0, 0, 0, 0, 0x000000, 0xFFFFFF},  // user input
    {STYLE_STDOUT, SETUPSTYLE_ALL, "Courier", 10, 0, 0, 0, 0, 0, 0x008000, 0xFFFFFF},  // StdOut
    {STYLE_STDERR, SETUPSTYLE_ALL, "Courier", 10, 0, 0, 0, 0, 0, 0x2020FF, 0xFFFFFF},  // StdErr
    {          27, SETUPSTYLE_ALL, "Verdana", 10, 0, 0, 0, 0, 0, 0x808080, 0xFFFFFF},
    {          28, SETUPSTYLE_ALL, "Verdana", 10, 1, 0, 0, 0, 0, 0x808000, 0xFFFFFF},
    {          29, SETUPSTYLE_ALL, "Verdana", 10, 0, 0, 1, 0, 1, 0xFF6060, 0xFFFFFF},
    {          30, SETUPSTYLE_ALL, "Verdana", 10, 1, 0, 0, 0, 0, 0xF0B450, 0xFFFFFF},
    {          31, SETUPSTYLE_ALL, "Verdana", 10, 0, 0, 0, 0, 0, 0x000000, 0xFFFFFF}
};

static STYLED_TEXT AboutMsg[] =
{
    {"CalcLuator ", 30}, {VERSION_STRING " ", 31},
    {" - Lua Calculator for Hiew\r\n", 28},
    {"Copyright (c) ", 31}, {"FSC", 30}, {", 2007-2009. ", 31}, {"http://www.fsc-soft.com", 29}, {", ", 31}, {DEVELOPERS_EMAIL, 29},  {"\r\n\r\n", 27},
    {"Press F5 to execute current line (or current selection if there is any)\r\n", 27},
    {"Press Ctrl+Break to abort script execution\r\n", 27},
    {"Press Esc or execute hiew.exit() to exit to Hiew\r\n", 27},
    {"Press F1 to show this short help message\r\n\r\n", 27},
    {"The following extra features are available for your scripts:\r\n", 27},
    {"Hiew control:\t\thiew.setoffset(lo[, hi]), hiew.setreload(), hiew.setmode(mode), hiew.exit()\r\n", 27},
    {"Hiew info:\t\thiew.filename, hiew.filesizelo, hiew.filesizehi, hiew.offsetlo, hiew.offsethi\r\n", 27},
    {"Hiew constants:\thiew.TEXT, hiew.HEX, hiew.CODE\r\n\r\n", 27},
    {"require() search order (! means path of Hiew executable):\r\n", 27},
    {"LUA_PATH_DEFAULT:\t" LUA_PATH_DEFAULT "\r\n", 27},
    {"LUA_CPATH_DEFAULT:\t" LUA_CPATH_DEFAULT "\r\n\r\n", 27},
    {"You can put any definitions you want in the %HIEWDIR%\\LuaScripts\\CalcLuatorInit.lua file\r\n", 27},
    {"which gets loaded automatically at startup\r\n\r\n", 27}
};

static void ShowHelp()
{
    if(SciSend(SCI_GETCURRENTPOS)) CrLf();
    DEFINETEXTS(AllMessages, AboutMsg);
    SciViewCmd.pTexts = &AllMessages;
    CallSciView(SCIVIEW_ADDTEXTS);
    SciSend(SCI_SCROLLCARET);
}

static void SetupSciView()
{
    /* setup indentation which may be left in an unknown state by other plugins */
    SciViewCmd.IntValue = 0;
    CallSciView(SCIVIEW_SETINDENT);

    /* setup styles 0 (StdOut) and 1 (StdErr) */
    DEFINESTYLES(AllStyles, LuaStyles);
    SciViewCmd.pStyles = &AllStyles;
    CallSciView(SCIVIEW_SETUPSTYLES);

    /* subclass Scintilla window so that we can intercept its messages */
    SciSend(SCI_SETMESSAGEPREVIEW, 0, MessagePreview);

    /* Load last .sav file (if any) */
    if(!LoadSav()) ShowHelp();
    SciSend(SCI_SCROLLCARET);

    /* load default initialization script */
    char ScriptName[MAX_PATH];
    GetModuleFileName(0, ScriptName, sizeof(ScriptName));
    PathRemoveFileSpec(ScriptName);
    strcat(ScriptName, "\\LuaScripts\\CalcLuatorInit.lua");
    if(PathFileExists(ScriptName)) ExecuteFile(ScriptName, 0, 0);
}

//------------------------------------------------------------------------
// WM_INITDIALOG message handler
//
static BOOL InitDialog()
{
    OutFlags = 0;

    /* get SciView information so that we can send messages to its Scintilla window */
    if(CallSciView(SCIVIEW_GETINFO) == HEM_OK)
    {
        DirectFn = SciViewCmd.GetInfo.DirectFn;
        DirectPtr = SciViewCmd.GetInfo.DirectPtr;
        hDialogWnd = SciViewCmd.GetInfo.hDialogWnd;
        hSciViewWnd = SciViewCmd.GetInfo.hSciViewWnd;
        if(DirectFn && DirectPtr)
        {
            /* Scintilla window has been successfully created by SciView so now we can do anything we want */
            SetupSciView();

            /* everything's fine - we can continue */
            return TRUE;
        }
    }
    
    /* something went wrong - time to close */
    CloseDialog(-2);
    return FALSE;
}

//------------------------------------------------------------------------
// WM_DESTROY message handler
//
static BOOL WmDestroy()
{
    CloseLua();

    /* save screen contents to .sav file. No error checking here because the operation is not vital */
    TEXTRANGE Tr;
    Tr.chrg.cpMin = 0;
    Tr.chrg.cpMax = SciSend(SCI_GETTEXTLENGTH);
    DWORD BufSize = 2 * Tr.chrg.cpMax + 2;
    Tr.lpstrText = (LPSTR)malloc(BufSize);
    if(Tr.lpstrText)
    {
        SciSend(SCI_GETSTYLEDTEXT, 0, &Tr);
        HANDLE hFile = CreateFile(SaveFileName, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
        if(hFile != INVALID_HANDLE_VALUE)
        {
            DWORD BytesWritten;
            /* note that we don't want to write the two trailing zeros added by SCI_GETSTYLEDTEXT */
            WriteFile(hFile, Tr.lpstrText, BufSize - 2, &BytesWritten, 0);
            CloseHandle(hFile);
        }
        free(Tr.lpstrText);
    }

    /* let the plugin process the message if it needs to */
    return FALSE;
}

//------------------------------------------------------------------------
// Get script and execute it
// If nothing is selected, executes the current line. Otherwise executes the selection
//
static void Execute(HWND hWnd, UINT NotifyMsg)
{
    TEXTRANGE Tr;
    Tr.lpstrText = 0;

    Tr.chrg.cpMin = SciSend(SCI_GETSELECTIONSTART);
    Tr.chrg.cpMax = SciSend(SCI_GETSELECTIONEND);

    if(Tr.chrg.cpMin == Tr.chrg.cpMax)
    {
        /* there is no selection, we will process the current line in this case */
        int CurLine = SciSend(SCI_LINEFROMPOSITION, SciSend(SCI_GETCURRENTPOS));

        /* unlike SCI_GETCURRENTLINE, this does not return any EOL characters */
        Tr.chrg.cpMin = SciSend(SCI_POSITIONFROMLINE, CurLine);
        Tr.chrg.cpMax = SciSend(SCI_GETLINEENDPOSITION, CurLine);
    }

    /* do nothing if the line is empty */
    if(Tr.chrg.cpMin == Tr.chrg.cpMax) return;

    /* allocate memory for script text including terminating 0 */
    Tr.lpstrText = (LPSTR)malloc(Tr.chrg.cpMax - Tr.chrg.cpMin + 1);
    if(Tr.lpstrText == 0)
    {
        PrintStdErr("Not enough memory");
        return;
    }

    /* retrieve text. SCI_GETTEXTRANGE returns the length of the retrieved text */
    if(SciSend(SCI_GETTEXTRANGE, 0, &Tr))
    {
        /* move the cursor past the last character of the selection or current line */
        SciSend(SCI_GOTOPOS, Tr.chrg.cpMax);
        CrLf();
        ExecuteScript(Tr.lpstrText, hWnd, WM_SCRIPTDONE);
    }

    /* since the chunk is now compiled, we can free allocated memory */
    free(Tr.lpstrText);
}

//------------------------------------------------------------------------
// WM_KEYDOWN message handler
// Returns TRUE if this is one of expected keys
//
#define keyCTRL         0x80000000
#define keyALT          0x40000000
#define keySHIFT        0x20000000

static BOOL WmKeyDown(HWND hWnd, DWORD Vk)
{
    /* in fact, virtual key code is byte-wide so we can use its upper bits for our own flags */
    Vk &= ~(keyCTRL | keyALT | keySHIFT);
    
    if(GetKeyState(VK_CONTROL) < 0) Vk |= keyCTRL;
    if(GetKeyState(VK_MENU)    < 0) Vk |= keyALT;
    if(GetKeyState(VK_SHIFT)   < 0) Vk |= keySHIFT;

    int Cmd;
    switch(Vk)
    {
        case VK_F5:
            Execute(hWnd, WM_SCRIPTDONE);
            return TRUE;

        case VK_F1:
            ShowHelp();
            return TRUE;

        case VK_CANCEL | keyCTRL:
            TerminateScript();
            return TRUE;

        /* QuickType keys */
        case 'P' | keyCTRL:         Cmd = ID_QUICKTYPE_PRINT;       break;
        case 'I' | keyCTRL:         Cmd = ID_QUICKTYPE_IFELSETHEN;  break;
        case 'F' | keyCTRL:         Cmd = ID_QUICKTYPE_FOR;         break;
        case 'W' | keyCTRL:         Cmd = ID_QUICKTYPE_WHILE;       break;
        case 'R' | keyCTRL:         Cmd = ID_QUICKTYPE_REPEAT;      break;
        case 'D' | keyCTRL:         Cmd = ID_QUICKTYPE_DO;          break;
        case 'A' | keyALT:          Cmd = ID_QUICKTYPE_ASSERT;      break;
        case 'F' | keyALT:          Cmd = ID_QUICKTYPE_FUNCTION;    break;
        case 'L' | keyCTRL:         Cmd = ID_QUICKTYPE_LOCAL;       break;

        default:
            /* this is not our key */
            return FALSE;
    }
    SendMessage(hWnd, WM_COMMAND, Cmd, 0);
    return TRUE;
}

//------------------------------------------------------------------------
// Insert N spaces
//
static void InsertSpaces(int N)
{
    if(N > 0 && N < 256)
    {
        char Buffer[256];
        memset(Buffer, ' ', N);
        SciSend(SCI_ADDTEXT, N, Buffer);
    }
}

//------------------------------------------------------------------------
// Type given expression
// Note: no buffer overflow control here so please check that input string does not contain long lines
//
static BOOL QuickType(LPCTSTR pExpr, int DeltaPos = -1)
{
    int CurPos = SciSend(SCI_GETCURRENTPOS);
    int CurLine = SciSend(SCI_LINEFROMPOSITION, CurPos);
    int LineStartPos = SciSend(SCI_POSITIONFROMLINE, CurLine);
    int Indent = CurPos - LineStartPos;

    char Buffer[512];
    PCHAR pOut = Buffer;

    while(true)
    {
        char Ch = *pExpr++;
        switch(Ch)
        {
            case 0:
                if(pOut > Buffer) SciSend(SCI_ADDTEXT, pOut - Buffer, Buffer);
                if(DeltaPos != -1) SciSend(SCI_GOTOPOS, CurPos + DeltaPos);
                return TRUE;

            case '\r':
                *pOut++ = '\r';
                *pOut++ = '\n';
                SciSend(SCI_ADDTEXT, pOut - Buffer, Buffer);
                InsertSpaces(Indent);
                pOut = Buffer;
                break;

            default:
                *pOut++ = Ch;
        }
    }
}

//------------------------------------------------------------------------
// WM_CONTEXTMENU message handler
//
static BOOL WmContextMenu(HWND hWnd)
{
    POINT Pt;
    GetCursorPos(&Pt);
    TrackPopupMenu(GetSubMenu(hMainMenu, 0), TPM_RIGHTBUTTON, Pt.x, Pt.y, 0, hWnd, 0);
    return TRUE;
}

//------------------------------------------------------------------------
// WM_COMMAND message handler
//
static BOOL WmCommand(HWND hWnd, int Cmd)
{
    switch(Cmd)
    {
        case ID_EXECUTE:        Execute(hWnd, WM_SCRIPTDONE);       break;
        case ID_ABORTSCRIPT:    TerminateScript();                  break;
        case ID_SHOWHELP:       ShowHelp();                         break;
        case ID_CLEAR:          CallSciView(SCIVIEW_CLEAR);         break;
        case ID_EXIT:           CloseDialog(0);                     break;
        case IDC_EXITTOHIEW:    CloseDialog(IDC_EXITTOHIEW);        break;

        case ID_QUICKTYPE_PRINT:        return QuickType("print()", 6);
        case ID_QUICKTYPE_IFELSETHEN:   return QuickType("if then\relse\rend", 3);
        case ID_QUICKTYPE_FOR:          return QuickType("for i = 1, 10, 2 do\rend", 16);
        case ID_QUICKTYPE_WHILE:        return QuickType("while do\rend", 6);
        case ID_QUICKTYPE_REPEAT:       return QuickType("repeat\runtil", 6);
        case ID_QUICKTYPE_DO:           return QuickType("do\rend", 2);
        case ID_QUICKTYPE_ASSERT:       return QuickType("assert()", 7);
        case ID_QUICKTYPE_FUNCTION:     return QuickType("function ()\rend", 9);
        case ID_QUICKTYPE_LOCAL:        return QuickType("local ");

        default:                return FALSE;
    }
    return TRUE;
}

//------------------------------------------------------------------------
// Retrieve and process URI
// Note: no checks for pOutBuffer overflow here so output buffer should have sufficient length.
//
static bool GetURI(PCHAR pOutBuffer, int Pos)
{
    char Temp[512];
    
    SciViewCmd.Url.Pos = Pos;
    SciViewCmd.Url.pOutText = 0;            // get required buffer size
    
    /* check if the string will fit our buffer */
    if(CallSciView(SCIVIEW_GETURL) < sizeof(Temp))
    {
        /* yeah, we can retrieve the URI */
        SciViewCmd.Url.pOutText = Temp;
        
        if(CallSciView(SCIVIEW_GETURL))
        {
            /* is this an e-mail address? */
            if(strchr(Temp, '@'))
            {
                /* is this a very special e-mail address? */
                PCHAR pSubj = strstr(Temp, DEVELOPERS_EMAIL) ? "?subject=[CalcLuator]" : "";

                /* construct e-mail address */
                sprintf(pOutBuffer, "mailto:%s%s", Temp, pSubj);

            }
            else strcpy(pOutBuffer, Temp);
            return true;
        }
    }
    return false;
}

//------------------------------------------------------------------------
// WM_NOTIFY message handler
//
static BOOL WmNotify(SCNotification *pNotify)
{
    /* we want to process notifications coming only from SciView window */
    if(pNotify->nmhdr.hwndFrom == hSciViewWnd)
    {
        /* we want to process only hot spot clicks */
        if(pNotify->nmhdr.code == SCN_HOTSPOTCLICK)
        {
            char URLBuffer[1024];

            if(GetURI(URLBuffer, pNotify->position))
                ShellExecute(0, 0, URLBuffer, 0, 0, SW_SHOWNORMAL);
        }
    }
    /* let the main dialog procedure process this message too */
    return FALSE;
}

//------------------------------------------------------------------------
// Our own dialog procedure
//
static BOOL CALLBACK DialogProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
    switch(uMsg)
    {
        case WM_INITDIALOG:     return InitDialog();
        case WM_DESTROY:        return WmDestroy();

        case WM_SYSKEYDOWN:
        case WM_KEYDOWN:        return WmKeyDown(hWnd, (DWORD)wParam);

        case WM_CONTEXTMENU:    return WmContextMenu(hWnd);
        case WM_COMMAND:        return WmCommand(hWnd, wParam);
        case WM_NOTIFY:         return WmNotify((SCNotification *)lParam);
        case WM_SCRIPTDONE:     return TRUE;            // we don't actually process this message and it is here just for demonstration purposes
    }

    /* let SciView process all other messages */
    return FALSE;
}

//------------------------------------------------------------------------
// Show SciView window
//
int ShowSciView()
{
    SciViewCmd.Show.DialogProc = DialogProc;
    return CallSciView(SCIVIEW_SHOW);
}

//------------------------------------------------------------------------
// StdOut and StdErr print functions
//
void __stdcall PrintStdOut(LPCTSTR pText)
{
    SciViewCmd.IntValue = STYLE_STDOUT;
    CallSciView(SCIVIEW_SETSTYLE);
    SciViewCmd.pText = pText;
    CallSciView(SCIVIEW_ADDTEXT);
}

void __stdcall PrintStdErr(LPCTSTR pText)
{
    SciViewCmd.IntValue = STYLE_STDERR;
    CallSciView(SCIVIEW_SETSTYLE);
    SciViewCmd.pText = pText;
    CallSciView(SCIVIEW_ADDTEXT);
}
